/*
 *  DEXID
 *	dsql.3
 *
 *  NAME
 *	dsql - Interbase dynamic SQL interface
 *
 *  SYNOPSIS
 *	#include <dsql.h>
 *	
 *	int q_init()
 *	
 *	int q_open_db(status, db, path)
 *	long *status;
 *	Q_db_handle *db;
 *	char *path;
 *	
 *	int q_close_db(status, db)
 *	long *status;
 *	Q_db_handle *db;
 *	
 *	int q_start_trans(va_alist)
 *	va_dcl
 *	
 *	int q_gen(status, q, sql, db, tx)
 *	long *status;
 *	Query *q;
 *	char *sql;
 *	Q_db_handle db;
 *	Q_tx_handle tx;
 *	
 *	int q_open_cursor(status, q, tx)
 *	long *status;
 *	Query *q;
 *	Q_tx_handle tx;
 *	
 *	int q_execute(status, q, tx)
 *	long *status;
 *	Query *q;
 *	Q_tx_handle tx;
 *	
 *	int q_fetch(status, q, ptrs)
 *	long *status;
 *	Query *q;
 *	QField *ptrs;
 *	
 *	int q_close_cursor(status, q)
 *	long *status;
 *	Query *q;
 *	
 *	int q_commit_trans(status, tx)
 *	long *status;
 *	Q_tx_handle *tx;
 *
 *	int q_rollback_trans(status, tx)
 *	long *status;
 *	Q_tx_handle *tx;
 *	
 *	int q_format_fld(buf, fmt, fptr)
 *	char buf[];
 *	char *fmt;
 *	QField *fptr;
 *
 *	int q_error_text(status, buf, sepstring)
 *	long *status;
 *	char *buf, *sepstring;
 *
 *
 *	int Q_FNULL(fptr)
 *	QField *fptr;
 *
 *	int Q_FSIZE(fptr)
 *	QField *fptr;
 *
 *	Data_Type Q_FTYPE(fptr)
 *	QField *fptr;
 *
 *	char * Q_CPV(fptr)
 *	QField *fptr;
 *
 *	short Q_SHRV(fptr)
 *	QField *fptr;
 *	
 *	int Q_INTV(fptr)
 *	QField *fptr;
 *
 *	float Q_FLTV(fptr)
 *	QField *fptr;
 *
 *	double Q_DBLV(fptr)
 *	QField *fptr;
 *
 *	struct tm Q_TMV(fptr)
 *	QField *fptr;
 * 
 *
 *	int Q_QTYPE(q)
 *	Query *q;
 *
 *	int Q_NUMCOLS(q)
 *	Query *q;
 *
 *
 *  DESCRIPTION
 *	dsql implements a C API interface to Interbase dynamic SQL.  Here are
 *	the major features:
 *	.sp 1
 *	Applications may interact with Interbase DBs via DSQL without
 *	precompilation.
 *	.sp 1
 *	"Base" Interbase DSQL functionality is extended -- multiple dbs,
 *	queries, and transaction contexts may be managed simultaneously.
 *	.sp 1
 *	Resources necessary to properly fetch data (e.g. buffers) are
 *	automatically created and destroyed; apps are insulated from this code
 *	overhead and are thus more reliable and robust.
 *
 *	q_init() must be called before any other function.  All Interbase SQL
 *	statements must be executed within a valid DB and transaction context.
 *	The DB context is created via q_open_db(); db_ will be populated with
 *	a non-zero handle upon success.  A transaction context is started with
 *	q_start_trans().  Currently, only the default consistent mode is
 *	available.  tx_ will be populated with a non-zero handle upon success.
 *
 *	q_gen() takes the db_ and tx_ handles along with a SQL statement in
 *	sql_ and populates a query structure q_ upon success.  The query
 *	structure can then be used in q_execute() if it is NOT a select type
 *	statement or q_open_cursor() if it is a select type statement.  The
 *	query type can be identified using Q_QTYPE(), which will return either
 *	Q_SELECT_TYPE or Q_EXECUTE_TYPE.  Note that the tx_ handle must once
 *	again be passed in either case to properly track lock status between
 *	the time the query was created and the time is was executed/opened.
 *	.sp 1
 *	Query structure resources will be deallocated after q_execute().  You
 *	must call q_gen() again to create a new query structure.
 *	.sp 1
 *	If the SQL statement in the query structure is a select-type, then one
 *	may call q_fetch() after opening the cursor with q_open_cursor().  The
 *	third argument to q_fetch() is an array of QField structures.  One
 *	QField structure is needed for each column in the select statement.
 *	The number of columns that will be fetched can be determined by using
 *	Q_NUMCOLS().  An efficient (and safe) way to ensure that you have only
 *	as many QField structures as you need is to dynamically allocate space
 *	for the QField structures based on the number returned by Q_NUMCOLS():
 *
 *	.ft C
 *	.nf
 *	    Query q;
 *	    QField *fptr;
 *	    ...
 *	    q_gen(status, &q, sql, dbhandle, txhandle);
 *	    fptr = (QField *) malloc(Q_NUMCOLS(&q) * sizeof(QField));
 *	    while((rc = q_fetch(status, &q, fptr)) != Q_EOC) {
 *	        ...
 *	    }
 *	.fi
 *	.ft R
 *
 *	Information about each QField structure can be retrieved using the Q_
 *	family of macros.  Typically, Q_FNULL() is called first to check if
 *	the field is null.  If Q_FNULL() indicates the field is \fInot\fR null
 *	by returning 0, then Q_FTYPE() may be used to determine the basic type
 *	of the data where where \fIData_Type\fP is an enumerated type, one of
 *	the following:
 *
 *	.ft C
 *	.nf
 *	    CHAR_TYPE	- always space-padded to length of field
 *	    VARY_TYPE	- returns EXACTLY what was stored, nulls 'n' all
 *	    SHORT_TYPE	- short int, 2 bytes
 *	    LONG_TYPE	- long int, 4 bytes
 *	    FLOAT_TYPE	- short floating, 4 bytes (not recommended)
 *	    DOUBLE_TYPE	- long floating, 8 bytes
 *	    DATE_TYPE	- quad int; no external representation
 *	    TIME_TYPE	- quad int; no external representation
 *	    TIMESTAMP_TYPE - quad int; no external representation
 *	    BLOB_TYPE	- quad int; no external representation
 *	    ARRAY_TYPE	- quad int; no external representation
 *	.fi
 *	.ft R
 *
 *	Additionally, Q_FSIZE() may be used to return the size of the column.
 *	Q_FSIZE() will return the string length of CHAR and VARY types and the
 *	sizeof() the other types.  The appropriate Q_*V() macro may then be
 *	used to fetch the datum from the structure.  NOTE: Interbase does
 *	\fBnot\fP populate the \fCtm_isdst\fR, \fCtm_zone\fR, or
 *	\fCtm_gmtroff\fR elements of the tm structure returned by Q_TMV().
 *	Also, the \fCtm_sec\fR, \fCtm_min\fR, and \fCtm_hour\fR elements may
 *	be set to zero or not depending on whether GDML or SQL was used to
 *	store the field -- it is probably best to avoid using them.
 *
 *	q_format_fld() provides an alternate means to format the data in fptr_
 *	into buffer buf_ as indicated by the fmt_ argument.  fmt_ may be any
 *	valid sprintf() format argument for CHAR, VARY, SHORT, LONG, FLOAT, or
 *	DOUBLE types; any valid strftime() format argument for DATE types; or
 *	the NULL pointer.  If fmt_ is the NULL pointer, then default
 *	formatting (e.g. "%lf" for doubles) is performed with the following
 *	exceptions:
 *	.sp 1
 *	.ft C
 *	.nf
 *	    DATE_TYPE	- MM/DD/YYYY
 *	    TIME_TYPE	- (not supported yet)
 *	    TIMESTAMP_TYPE - (not supported yet)
 *	.fi
 *	.ft R
 *	.sp 1
 *	
 *	BLOB_TYPE and ARRAY_TYPE fields have no default formatting and
 *	q_format_fld() will essentially be a no-op if called with fields of
 *	this type.
 *	.sp 1
 *	When q_fetch() returns Q_EOC, the cursor has been exhausted and may be
 *	be closed with q_close_cursor().  Query structure resources are
 *	deallocated at this time in the same way as q_execute().
 *	.sp 1
 *	At any point, the application can commit or rollback work with
 *	q_commit_trans() and q_rollback_trans(), respectively.  The
 *	transaction context tx_ will become undefined at this time.  Any
 *	further calls to any function requiring a transaction handle must
 *	first call q_start_trans() to  re-establish a valid transaction
 *	context.
 *	.sp 1
 *	The application disconnects from the database using the q_close_db()
 *	call.  It is important to check the error status vector after this
 *	call because Interbase will \fBNOT\fR close a DB that has active
 *	transactions.  All transactions must be committed or rolled back
 *	before q_close_db() is issued.
 *	.sp 1
 *	If an error occurs, the error text as generated by the DB engine can
 *	be fetched via q_error_text().  The buf_ is filled with descriptive
 *	text strings separated by sepstring_ based on the status information
 *	present in the status_ vector.  The number of strings placed in buf_
 *	is returned.
 *
 *
 *  NOTES
 *	.nf
 *	All functions except q_fetch return 0, 1, or -1
 *	0  - SUCCESS
 *	1  - Interbase internal failure.  Examine the status vector for more 
 *	     details.
 *	-1 - dsql package failure.   Examine the status vector for more 
 *	     details.
 *	q_fetch() returns -2 (Q_EOC) upon reaching end of cursor.
 *	.fi
 *	.sp 1
 *	It is \fBnot\fR a bug/mistake/design flaw that q_gen() requires a
 *	valid transaction handle.  Queries on the Interbase DB data dictionary
 *	itself are under transaction control just like regular queries, so a
 *	tx handle must be supplied.
 *
 *  BUGS
 *	No runtime modification of transaction context type -- it's always
 *	consistent read/write.
 *	.sp 1
 *	Perhaps not a bug, but there is no way to portably create varargs
 *	lists for terminal functions that themselves are not set up with a 
 *	vararg interface.  Specifically, isc_start_transaction() is a function
 *	with the following prototype:
 *	.sp 1
 *	.ft C
 *	.nf
 *		isc_start_transaction(long *status, int n, ...);
 *	.fi
 *	.ft R
 *	.sp 1
 *	where the vararg list is n instances of a DB descriptor clumplet.  The
 *	descriptor clumplet is:
 *	.sp 1
 *	.ft C
 *	.nf
 *		long *dbhandle, (short) sizeof(db_param), char *db_param
 *	.fi
 *	.ft R
 *	.sp 1
 *	Since I can't pass:
 *	.sp 1
 *	.ft C
 *	.nf
 *		isc_start_transaction(status, n, va_alist);
 *	.fi
 *	.ft R
 *	.sp 1
 *	or equiv, we have to...grrrr...code up n calls to the function.
 *
 *  COPYRIGHT
 *	Copyright 1993 Buzz Moschetti
 *
 *	Permission to use, copy, modify, distribute, and sell this
 *	software and its documentation for any purpose is hereby granted
 *	by Buzz Moschetti (the "Author") without fee, provided that the
 *	above copyright notice appear in all copies and that both that
 *	copyright notice and this permission notice appear in supporting
 *	documentation, and that the name of the Author. or contractions
 *	or abbreviations of the same not be used in advertising or
 *	publicity pertaining to distribution of the software without
 *	prior specific, written permission.  The Author does not make
 *	any representations about the suitability of this software for
 *	any purpose.  It is provided "as is" without express or implied
 *	warranty.
 *
 *	THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *	SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 *	AND FITNESS, IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 *	SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *	WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 *	IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *	ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *	THIS SOFTWARE.
 *	 
 *  AUTHOR
 *	Buzz Moschetti
 *
 *
 */
/* LINTLIBRARY */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <varargs.h>


#include "dsql.h"


static Alignptr qmalloc();
static Data_Type sql2dtype();
static int genfetchbuffer();
static int get_qid();
static void release_qid();
static void q_store_err();


#define ALIGN(n,b) ((n + b - 1) & ~(b - 1))

typedef struct vary {
      short    vary_length;
      char    vary_string [1];
} VARY;

/*
 *  SQL is PREPAREd into this template area.  Then, sqlda->sqld is checked to
 *  see how many items were selected.  A NEW sqlda is then qmalloced with only
 *  that many "slots" in the sqlvar and the info is copied over.
 */
static SQLDA *tsqlda;







int
q_open_db(status, db, path)
long *status;
Q_db_handle *db;
char *path;
{
      /* Force resetting of db handle... */
      *db = 0;

      (void) isc_attach_database (status, 0, path, db, 0, 0);

      return status[1];
}


int
q_close_db(status, db)
long *status;
Q_db_handle *db;
{
      (void) isc_detach_database(status, db);

      return status[1];
}
	    


int
q_start_trans(va_alist)
va_dcl
{
      struct {
	    int *dbptr;
	    short size;
	    char tpb;
      } args[MAXOPENDBS];

      int dbhandles[MAXOPENDBS];

      long *status;
      gds_tr_handle *tx;
      int i, j;
      short dbcount;
      va_list ap;



      va_start(ap);


      /* Strip the required args off ap: */
      status  = va_arg(ap, long *);
      tx      = va_arg(ap, gds_tr_handle *);
      dbcount = (short) va_arg(ap, int);  

      /* Force resetting of tx handle... */
      *tx = 0;


      i = dbcount;
      j = 0;

      while(i--) {
	    dbhandles[j] = va_arg(ap, int);  /* Store the handle...*/

	    args[j].dbptr = &dbhandles[j];

	    /* For now, sizeof(tx param block) is 0.  No flags will be
	     * recognized:
	     */
	    args[j].size = 0;
	    args[j].tpb  = 0;

	    /* Eat the next arg, the flags arg... */
	    (void) va_arg(ap, int);

	    j++;
      }

      /*  This is the ugliest thing EVER: */
      switch(dbcount) {
	  case 1:
	    isc_start_transaction (status, tx, dbcount,
				   args[0].dbptr,
				   args[0].size,
				   args[0].tpb);
	    break;

	  case 2:
	    isc_start_transaction (status, tx, dbcount,
				   args[0].dbptr,
				   args[0].size,
				   args[0].tpb,
				   args[1].dbptr,
				   args[1].size,
				   args[1].tpb);
	    break;

	  case 3:
	    isc_start_transaction (status, tx, dbcount,
				   args[0].dbptr,
				   args[0].size,
				   args[0].tpb,
				   args[1].dbptr,
				   args[1].size,
				   args[1].tpb,
				   args[2].dbptr,
				   args[2].size,
				   args[2].tpb);
	    break;

	  case 4:
	    isc_start_transaction (status, tx, dbcount,
				   args[0].dbptr,
				   args[0].size,
				   args[0].tpb,
				   args[1].dbptr,
				   args[1].size,
				   args[1].tpb,
				   args[2].dbptr,
				   args[2].size,
				   args[2].tpb,
				   args[3].dbptr,
				   args[3].size,
				   args[3].tpb);
	    break;

	  case 5:
	    isc_start_transaction (status, tx, dbcount,
				   args[0].dbptr,
				   args[0].size,
				   args[0].tpb,
				   args[1].dbptr,
				   args[1].size,
				   args[1].tpb,
				   args[2].dbptr,
				   args[2].size,
				   args[2].tpb,
				   args[3].dbptr,
				   args[3].size,
				   args[3].tpb,
				   args[4].dbptr,
				   args[4].size,
				   args[4].tpb);
	    break;

	  case 6:
	    isc_start_transaction (status, tx, dbcount,
				   args[0].dbptr,
				   args[0].size,
				   args[0].tpb,
				   args[1].dbptr,
				   args[1].size,
				   args[1].tpb,
				   args[2].dbptr,
				   args[2].size,
				   args[2].tpb,
				   args[3].dbptr,
				   args[3].size,
				   args[3].tpb,
				   args[4].dbptr,
				   args[4].size,
				   args[4].tpb,
				   args[5].dbptr,
				   args[5].size,
				   args[5].tpb);
	    break;
      }

      va_end(ap);

      return status[1];
}


int
q_gen(status, q, sql, db, tx)
long *status;
Query *q;
char *sql;
Q_db_handle db;
Q_tx_handle tx;
{
      q->qnum = get_qid();

      /* 
       *  Yes - the space IS NECESSARY since query & cursor descriptors in
       *  Interbase are space-terminated strings, not null-terminated  :-(
       */
      (void) sprintf(q->qid, "Q%03d ", q->qnum);

      /* Even though it may seem like isc_prepare() is going to modify
       * the contents of db and tx, in fact the OSRI manual says they
       * are input values ONLY; thus, we can take the address here in the
       * function and not incur the headache of carting the ref address
       * in the q_gen() call iteself.
       */
      isc_prepare (status,
		   (gds_db_handle *) &db,
		   (gds_tr_handle *) &tx,
		   q->qid, 0, sql, tsqlda);

      if(isc_sqlcode(status) != 0) {
	    /* Release the Q num back to the pool: */
	    release_qid(q->qnum);

	    return status[1];
      }


      q->num_recs = 0;


      /*
       *  Gen a NEW copy of the descriptor area:
       */
      q->sqlda = (SQLDA *) qmalloc(SQLDA_LENGTH(tsqlda->sqld));
      (void) memcpy((void *) q->sqlda, (void *) tsqlda, SQLDA_LENGTH(tsqlda->sqld));
      q->sqlda->sqln = tsqlda->sqld;


      /* If no fields are to be fetched (slqda->sqld == 0) then this is 
       * non-select type statement.  Non-selects do NOT get fetch buffers
       * or cursor resources genereated via isc_declare()
       * execute the statement and leave:
       */
      if (!tsqlda->sqld) {
	    q->query_type = Q_EXECUTE_TYPE;   /* delete, update, insert */
	    q->cursor_status = -1;
	    q->buffer = (char *)0;  /* no mistake about it... */
	    q->buffer_len = 0;
	    q->transfer_len = 0;
      } else {
	    q->query_type = Q_SELECT_TYPE;   /* select */
	    (void) genfetchbuffer(q);
	    isc_declare(status, q->qid, q->qid);
	    q->cursor_status = 0;
      }

      return status[1];
}


int
q_open_cursor(status, q, tx)
long *status;
Query *q;
Q_tx_handle tx;
{
      /* Make sure this is a SELECT type query: */
      if(q->query_type != Q_SELECT_TYPE) {
	    q_store_err(status, 
			"error opening cursor",
			q->qid,
			"statement is not select type",
			0);
	    return -1;
      }


      /*  The (SQLDA *)0 means no input SQLDA is to be used; that is,
       *  there are no ? variable parameters in the SQL query.
       */
      isc_open(status, (gds_tr_handle * ) &tx, q->qid, (SQLDA *)0);

      if(!status[1]) {
	    q->cursor_status = 1;
      }

      return status[1];
}



int
q_execute(status, q, tx)
long *status;
Query *q;
Q_tx_handle tx;
{
      /* Make sure this is NOT a SELECT type query: */
      if(q->query_type == Q_SELECT_TYPE) {
	    q_store_err(status, 
			"error executing statement",
			q->qid,
			"statement is a select type",
			0);
	    return -1;
      }


      /*  The (SQLDA *)0 means no input SQLDA is to be used; that is,
       *  there are no ? variable parameters in the SQL query.
       */
      isc_execute(status, (gds_tr_handle *) &tx, q->qid, (SQLDA*) 0);

      /* Release the Q num back to the pool: */
      release_qid(q->qnum);

      /* Free q resources. NO BUFFER TO free()! */
      free(q->sqlda);

      return status[1];
}



int
q_fetch(status, q, ptrs)
long *status;
Query *q;
QField *ptrs;
{
      SQLVAR *var;
      int rc, i;


      union {
	    VARY *vp;
	    char *cp;
	    short *sp;
	    int *ip;
	    double *dp;
	    float *fp;
      } CX;

      
      /* Make sure this is a SELECT type query: */
      if(q->query_type != Q_SELECT_TYPE) {
	    q_store_err(status, 
			"error fetching from cursor",
			q->qid,
			"statement is not select type",
			0);
	    return -1;
      }


      /*
       *  POPULATE the pointers in sqlda!
       */
      rc = isc_fetch(status, q->qid, q->sqlda);

      if(rc == 100) {
	    return Q_EOC;
      }

      if(isc_sqlcode(status) != 0) {
	    return status[1];
      }

      
      /*
       *  Whip through sqlda, assigning the incoming pointers to the var
       *  pointers set up in genfetchbuffer():
       */
      for(i = 0; i < q->sqlda->sqld; i++) {
	    
	    var = &q->sqlda->sqlvar[i];

	    ptrs[i].type = sql2dtype(var->sqltype & ~1);
	    ptrs[i].len  = var->sqllen;


	    CX.cp = var->sqldata;

	    switch(ptrs[i].type) {
	    /* NULL terminate as appropriate: */
		case CHAR_TYPE:
		  var->sqldata[var->sqllen] = '\0';
		  ptrs[i].v.cp = var->sqldata;
		  break;

		case VARY_TYPE:
		  ptrs[i].v.cp = CX.vp->vary_string;
		  ptrs[i].v.cp[CX.vp->vary_length] = '\0';
		  break;

		case SHORT_TYPE:
		  ptrs[i].v.s = *CX.sp;
		  break;

		case LONG_TYPE:
		  ptrs[i].v.i = *CX.ip;
		  break;

		case FLOAT_TYPE:
		  ptrs[i].v.flt = *CX.fp;
		  break;

		case DOUBLE_TYPE:
		  ptrs[i].v.dbl = *CX.dp;
		  break;

		case DATE_TYPE:
		  isc_decode_date(CX.cp, &ptrs[i].v.tm);
		  break;

		default:
		  ptrs[i].v.cp = var->sqldata;
		  break;
	    }


	    /*  Was a null value pulled (i.e. sqlind < 0) from a field that
	     *  supports NULLS (i.e. sqltype & 1 == true)?
	     */
	    if ((var->sqltype & 1) && (*var->sqlind < 0)) {
		  ptrs[i].null = (short) 1;
	    } else {
		  ptrs[i].null = (short) 0;
	    }
      }
	    
      q->num_recs++;

      return 0;
}


int
q_close_cursor(status, q)
long *status;
Query *q;
{
      /* Make sure this is a SELECT type query: */
      if(q->query_type != Q_SELECT_TYPE) {
	    q_store_err(status, 
			"error fetching from cursor",
			q->qid,
			"statement is not select type",
			0);
	    return -1;
      }

      isc_close(status, q->qid);

      if(!status[1]) {
	    q->cursor_status = 0;
      }

      /* Release the Q num back to the pool: */
      release_qid(q->qnum);

      /* Free q resources: */
      free(q->sqlda);
      free(q->buffer);

      return status[1];
}


int
q_commit_trans(status, tx)
long *status;
Q_tx_handle *tx;
{
      isc_commit_transaction(status, tx);

      return status[1];
}


int
q_rollback_trans(status, tx)
long *status;
Q_tx_handle *tx;
{
      isc_rollback_transaction(status, tx);

      return status[1];
}



int
q_format_fld(buf, fmt, fptr)
char *fmt;
char buf[];
QField *fptr;
{
      /* Was a null value pulled? */
      if(fptr->null) {
	    buf[0] = '\0';
      }

      /* Not a NULL value!  Get it into the xferbuf... */
      else {
	    switch(fptr->type) {
		case SHORT_TYPE:
		  (void) sprintf(buf, fmt ? fmt : "%hd", fptr->v.s);
		  break;
		  
		case LONG_TYPE:
		  (void) sprintf(buf, fmt ? fmt : "%ld", fptr->v.i);
		  break;
		  
		case FLOAT_TYPE:
		  (void) sprintf(buf, fmt ? fmt : "%f", fptr->v.flt);
		  break;
		  
		case DOUBLE_TYPE:
		  (void) sprintf(buf, fmt ? fmt : "%lf", fptr->v.dbl);
		  break;
		  
		case CHAR_TYPE:
		case VARY_TYPE:
		  (void) sprintf(buf, fmt ? fmt : "%s", fptr->v.cp);
		  break;
		  
		case DATE_TYPE:
		  if(fmt) {
			(void) strftime(buf, 512, fmt, &fptr->v.tm);
		  } else {
			(void) sprintf(buf, "%02d/%02d/%4d",
				       fptr->v.tm.tm_mon + 1,
				       fptr->v.tm.tm_mday,
				       fptr->v.tm.tm_year + 1900);
		  }
		  break;
	    }      
      }

      return 0;
}



int
q_error_text(status, buf, sepstring)
long *status;
char *buf, *sepstring;
{
      int i;
      long *v;

      v = status;
      i = 0;

      while(isc_interprete(buf, &v)) {		  
	    i++;
	    (void) strcat(buf, sepstring);
	    buf += strlen(buf);
      }

      return i;
}



      
      
      
/*
 *   Special dsql resource generation!
 */

#define   MAXQINDEX    32

static int qindex[MAXQINDEX];


int
q_init()
{
      tsqlda = (SQLDA *) qmalloc(SQLDA_LENGTH(SQLN));
      tsqlda->sqln = SQLN;

      (void) memset((void *) qindex, '\0', sizeof(qindex));

      return 0;
}


static struct sqlconv {
      Data_Type dtype;
      int native;
} sqlconv[] = {
      {   CHAR_TYPE, SQL_TEXT},
      {   VARY_TYPE, SQL_VARYING},
      {  SHORT_TYPE, SQL_SHORT},
      {   LONG_TYPE, SQL_LONG},
      {  FLOAT_TYPE, SQL_FLOAT},
      { DOUBLE_TYPE, SQL_DOUBLE},
      {   DATE_TYPE, SQL_DATE},

      {   TIME_TYPE, -1},
      {   BLOB_TYPE, -1},
      {  ARRAY_TYPE, -1},
      {TIMESTAMP_TYPE, -1}
};
#define  SQLCONVSIZE  (sizeof(sqlconv)/sizeof(sqlconv[0]))

static Data_Type
sql2dtype(n)
int n;
{
      int i;
      for(i = 0; i < SQLCONVSIZE; i++) {
	    if(sqlconv[i].native == n) {
		  return sqlconv[i].dtype;
	    }
      }

      return -1;
}



static int
get_qid()
{
      int i;


      for(i = 0; i < MAXQINDEX; i++) {
	    if(qindex[i] == 0) {
		  break;
	    }
      }

      qindex[i] = 1; /* in use */

      return i;
}

static void
release_qid(qnum)
int qnum;
{
      qindex[qnum] = 0; /* NOT in use */
}



/*  Use SHORTS to keep it as small as possible -- you'll see why... */
struct place_holder {
      short data;
      short nullind;
};
      
static int
genfetchbuffer(q)
Query *q;
{
      SQLVAR *var, *end;
      int i;

      short alignment, type, offset;

      /*
       *  This struct array holds the proper offsets into q->buffer for
       *  each column's data and null indicator.  We know that we cannot
       *  fetch more than SQLN columns (see dsql.h) so there is no need
       *  to dyn alloc the space.  SQLN * sizeof(struct place_holder) must
       *  be as small as possible to keep the stack push reasonable...
       */
      struct place_holder ph[SQLN];
      
      /* Init */
      q->buffer_len    = 0;
      q->transfer_len  = 0;


      for(offset = 0, i = 0, var = q->sqlda->sqlvar, end=var + q->sqlda->sqld;
	  var<end; var++, i++) {

	    /*
	     *  Alignment properties
	     */
	    alignment = var->sqllen;

	    type = var->sqltype & ~1;

	    if (type == SQL_TEXT) {
		  alignment=1;
	    }
	    else if (type == SQL_VARYING) {
		  alignment=sizeof(short);
	    }


	    /*
	     *  Data pointer assignment
	     */
	    offset = ALIGN (offset, alignment);
	    ph[i].data = offset;

	    /*
	     *  After data pointer advance
	     */
	    offset += var->sqllen;
	    q->transfer_len += var->sqllen;

	    /*  DSQL does NOT add in the fact that SQL_VARYING types need an
	     *  EXTRA 2 bytes for the length header.  Thus, a SQL_VARYING 
	     *  field of length, say, 12 needs 14 bytes!
	     */
	    if(type == SQL_VARYING) {
		  offset += 2;
		  q->transfer_len += 2;
	    }
	    
	    /*  Add 1 for trailing null if a string type.  Don't add to 
	     *  transfer_len, though...
	     */
	    if(type == SQL_TEXT || type == SQL_VARYING) {
		  offset++;
	    }


	    /*
	     *  Null indicator assignment
	     */
	    offset = ALIGN (offset, sizeof (short));
	    ph[i].nullind = offset;

	    /*
	     *  After data pointer advance
	     */
	    offset += sizeof(short);
	    q->transfer_len += sizeof(short);
      }


      /*  We know how big to make to buffer now: */
      q->buffer = malloc((int) offset);
      q->buffer_len = offset;


      /*
       *  Now whip through the sqlvars and assign pointers:
       */
      for(i = 0, var = q->sqlda->sqlvar, end=var + q->sqlda->sqld;
	  var<end; var++, i++) {
	    var->sqldata = (char *) q->buffer + ph[i].data;
	    var->sqlind  = (short *) ((char *) q->buffer + ph[i].nullind);
      }
      

      return 0;
}

static Alignptr
qmalloc(len)
int len;
{
      union {
	    Alignptr ap;
	    char *cp;
      } CX;

      CX.cp = malloc(len);

      return CX.ap;
}


static void
q_store_err(va_alist)
va_dcl
{
      long *status;
      va_list ap;

      va_start(ap);

      /* Strip the required args off ap: */
      status  = va_arg(ap, long *);

      /*
       *  Indicate an interperl error with status[1] = -1:
       */
      status[1] = -1;

      /* Advance status to first eligible text slot: */
      status += 2;

      /*
       *  Megacheating here.  We KNOW that sizeof(char *) and sizeof(long)
       *  are the same so we can store text pointers in the status vector.
       *  Interbase cheats this way, too, so this will NOT be "the"
       *  non-portable code fragment:
       */
      while ((*status++ = (long) va_arg(ap, char *)) != (long)0)
	    ;

      va_end(ap);
}


