/*
 *  DEXID
 *	interperl.c
 *
 *  SYNOPSIS
 *	&dbready($dbpath);
 *	&dbstarttrans($dbhandle1, $flags1 [, $dbhandle2, $flags2, ... ]);
 *	&dbprepare($sql_statement, $dbhandle, $txhandle);
 *	&dbopencursor($qhandle, $tx);
 *	&dbfetchrow($qhandle);
 *	&dbclosecursor($qhandle);
 *	&dbcommit($txhandle);
 *	&dbfinish($dbhandle);
 *	&dberrtxt;
 *	
 *  DESCRIPTION
 *	Interbase extensions to perl
 *	
 *	Design Overview
 *	This module mates the perl user defined function interface (Interperl)
 *	with my Interbase dsql package.  Basically:
 *
 *	Dsql is a C (i.e. NOT precompiled!) API to the Interbase DSQL engine. 
 *	Multiple databases may be managed via a uniform and reliable
 *	transaction and locking mechanism.  Resources are managed through
 *	integer handles and a special structure called the Query structure.
 *	All functions return 0 for success, 1 if an Interbase internal error
 *	has occurred, or -1 if a dsql package error has occurred -- see the
 *	dsql man page for more details.
 *
 *	The Interperl package provides all the same elemental functionality as
 *	the dsql API -- multiple DBs, transaction and cursor contexts, etc.
 *	The main difference is that this module interprets the error and
 *	handle return values and translates them into values that make
 *	programming in perl easier.  For example, the DSQL API function
 *	.nf
 *		    int q_open_db(status, &dbhandle, dbpath);
 *	.fi
 *	returns 0 on success and fills in &dbhandle.  The perl equivalent of
 *	this statement is:
 *	.nf
 *		    &dbready($dbpath);
 *	.fi
 *	which returns a scalar integer handle > 0 on success or zero upon
 *	failure, making it easy to utilize the following perl idiom:
 *	.nf
 *		    $dbhandle = &dbready($dbpath) || die "Couldn't open it.";
 *	.fi
 *	Of course, we lose the top level ability to distinguish between an
 *	internal Interbase error and a dsql package error, but in general we
 *	don't care about that in perl.
 *
 *  NOTES
 *	Notes on the perl stack:
 *	The stack is instantiated in the main modules and addressed with the
 *	macro st[n], where n is the nth item on the stack.
 *	
 *	Item 0, or st[0] is the return argument for scalars.
 *
 *	See case US_dbfetchrow for an example of how to return a list.
 *
 *	If the stack is "backed up" by doing (--sp), then the return value 
 *	is the null list.
 *
 *  	str_get(THE STACK)    - Returns a string
 *  	str_gnum(THE STACK)   - Returns an int.
 *
 *	str_numset(st[n], (double) n)  - Sets the stack.  n must
 *	a type that can be cast to (double).  Coercion unions must be used
 *	to return pointers and other non-castables.
 *
 *  BUGS
 *	The dberrtxt calls isc_interprete() directly to parse the status
 *	vector.
 *
 *  COPYRIGHT
 *	Copyright 1993 Buzz Moschetti
 *
 *	Permission to use, copy, modify, distribute, and sell this
 *	software and its documentation for any purpose is hereby granted
 *	by Buzz Moschetti (the "Author") without fee, provided that the
 *	above copyright notice appear in all copies and that both that
 *	copyright notice and this permission notice appear in supporting
 *	documentation, and that the name of the Author. or contractions
 *	or abbreviations of the same not be used in advertising or
 *	publicity pertaining to distribution of the software without
 *	prior specific, written permission.  The Author does not make
 *	any representations about the suitability of this software for
 *	any purpose.  It is provided "as is" without express or implied
 *	warranty.
 *
 *	THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *	SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 *	AND FITNESS, IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 *	SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *	WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 *	IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *	ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *	THIS SOFTWARE.
 *	 
 *  AUTHOR
 *	Buzz Moschetti
 *
 */

/* Required to interface with perl: */

#include "EXTERN.h" 
#include "perl.h"
#undef MAX
#undef MIN
#if !defined(VERSION3)
#define str_2static(s)		str_2mortal(s)
#endif

#include "dsql.h"


/* Local decs:  -not- required by perl, but will be installed
 * as pointers-to-functions.   You probably won't need to change add
 * or delete these unless you want the user subroutines to have
 * different entry points.
 */
static int usersub();
static int userset();
static int userval();



/* 
 *  User subroutines:
 *
 */
static enum usersubs
{
      US_dbready,
      US_dbfinish,
      US_dbstarttrans,
      US_dbcommit,
      US_dbrollback,
      US_dbprepare,
      US_dbexecute,
      US_dbopencursor,
      US_dbfetchrow,
      US_dbclosecursor,
      US_dberrtxt
};



/*  General purpose play buf: */
static char s[1024];

/*  The status vector: */
static long status[20];


/*  The error return value is zero: */
#define  ERROR_VALUE    0.0

/*  This is the return value for funcs which do not return handles: */
#define  SUCCESS_VALUE  1.0





/*
 *  E N T R Y   P O I N T
 *
 *  userinit() is called by the perl core upon startup.
 */
int userinit()
{
    struct ufuncs uf;
    char *filename = "interperl.c";

    uf.uf_set = userset;
    uf.uf_val = userval;

    /*  Set up the functions we need: */
    make_usub("dbready",	US_dbready,	usersub, filename);
    make_usub("dbfinish", 	US_dbfinish,	usersub, filename);

    make_usub("dbstarttrans", 	US_dbstarttrans,usersub, filename);
    make_usub("dbcommit", 	US_dbcommit,    usersub, filename);
    make_usub("dbrollback", 	US_dbrollback,  usersub, filename);
	
    make_usub("dbprepare",	US_dbprepare,	usersub, filename);
    make_usub("dbexecute",	US_dbexecute,	usersub, filename);
    make_usub("dbopencursor",	US_dbopencursor,usersub, filename);
    make_usub("dbfetchrow",	US_dbfetchrow,  usersub, filename);
    make_usub("dbclosecursor",	US_dbclosecursor,usersub, filename);

    make_usub("dberrtxt",	US_dberrtxt,	usersub, filename);


    /* dsql init */
    q_init();
}





static int
usersub(ix, sp, items)
int ix;
register int sp;
register int items;
{
      /* Work up the stack addressing variable: */
      STR **st = stack->ary_array + sp;
      ARRAY *ary = stack;    

      register int i, j, k;
      register char *t;

      int txhandle, num_dbs, rc;
      int dbhandle;
      int dbhandles[MAXOPENDBS];   /*  For tx  :-(  */


      /*  This is needed to effectively coerce the stack into holding 
       *  a (Query *) type.  This will NOT WORK on those
       *  machines where sizeof(Query *) > sizeof(double), but I'm assuming
       *  that we won't find any.   :-{
       */
      union {
	    Query *q;
	    double d;
      } qptr;
	
      /* Set up the max amount of ptrs... */
      QField fptrs[SQLN];




      /* Mega function dispatch: */
      switch (ix) {

	    /* &dbready(string "path"); */
	  case US_dbready:
	    if (items != 1) {
		  fatal("Usage: $dbhandle = &dbready($path)");
	    } else {

		  /* Get pathname off the stack: */
		  t = (char *) str_get(st[1]);
		  
		  /* Force handle val to zero... */
		  dbhandle = 0;
		  
		  if(q_open_db(status, &dbhandle, t)) {
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], (double) dbhandle);
		  }		  
	    }

	    break;



	    /* &dbfinish(int dbhandle); */
	  case US_dbfinish:
	    if(items != 1) {
		  fatal("Usage: $retcode = &dbfinish($dbhandle)");
	    } else {
  
		  dbhandle = (int) str_gnum (st[1]);
		  
		  if(q_close_db(status, &dbhandle)) {
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], SUCCESS_VALUE);
		  }
	    }
	    break;



	    /* &dbstarttrans( --- args --- ); */
	  case US_dbstarttrans:
	    if(items % 2) {
		  fatal("Usage: $retcode = &dbstarttrans($dbhandle1, $flags1 [, $dbhandle2, $flags2, ... ]");
	    }

	    num_dbs = items/2;


	    /* Force a new TX handle val to zero... */
	    txhandle = 0;

	    /*  This is HIDEOUS! */
	    switch(num_dbs) {
		case 1:
		  dbhandles[0] = (int) str_gnum (st[1]);
		  rc = q_start_trans(status, &txhandle, 1,
				     dbhandles[0], 0);
		  break;
		  
		case 2:
		  dbhandles[0] = (int) str_gnum (st[1]);
		  dbhandles[1] = (int) str_gnum (st[3]);

		  rc = q_start_trans(status, &txhandle, 2,
				     dbhandles[0], 0,
				     dbhandles[1], 0);
		  break;

		case 3:
		  dbhandles[0] = (int) str_gnum (st[1]);
		  dbhandles[1] = (int) str_gnum (st[3]);
		  dbhandles[2] = (int) str_gnum (st[5]);

		  rc = q_start_trans(status, &txhandle, 3,
				     dbhandles[0], 0,
				     dbhandles[1], 0,
				     dbhandles[2], 0);
		  break;

		case 4:
		  dbhandles[0] = (int) str_gnum (st[1]);
		  dbhandles[1] = (int) str_gnum (st[3]);
		  dbhandles[2] = (int) str_gnum (st[5]);
		  dbhandles[3] = (int) str_gnum (st[7]);

		  rc = q_start_trans(status, &txhandle, 4,
				     dbhandles[0], 0,
				     dbhandles[1], 0,
				     dbhandles[2], 0,
				     dbhandles[3], 0);
		  break;

		case 5:
		  dbhandles[0] = (int) str_gnum (st[1]);
		  dbhandles[1] = (int) str_gnum (st[3]);
		  dbhandles[2] = (int) str_gnum (st[5]);
		  dbhandles[3] = (int) str_gnum (st[7]);
		  dbhandles[4] = (int) str_gnum (st[9]);

		  rc = q_start_trans(status, &txhandle, 5,
				     dbhandles[0], 0,
				     dbhandles[1], 0,
				     dbhandles[2], 0,
				     dbhandles[3], 0,
				     dbhandles[4], 0);
		  break;

		case 6:
		  dbhandles[0] = (int) str_gnum (st[1]);
		  dbhandles[1] = (int) str_gnum (st[3]);
		  dbhandles[2] = (int) str_gnum (st[5]);
		  dbhandles[3] = (int) str_gnum (st[7]);
		  dbhandles[4] = (int) str_gnum (st[9]);
		  dbhandles[5] = (int) str_gnum (st[11]);

		  rc = q_start_trans(status, &txhandle, 6,
				     dbhandles[0], 0,
				     dbhandles[1], 0,
				     dbhandles[2], 0,
				     dbhandles[3], 0,
				     dbhandles[4], 0,
				     dbhandles[5], 0);
		  break;
	    }

	    if(rc) {
		  str_numset(st[0], ERROR_VALUE);
	    } else {
		  str_numset(st[0], (double) txhandle);
	    }

	    break;
	    


	  case US_dbprepare:
	    if(items != 3) {
		  fatal("Usage: $qhandle = &dbprepare($statement, $dbhandle, $txhandle)");
	    } else {

		  /* Get statement, dbhandle, and txhandle: */
		  t = (char *) str_get(st[1]);
		  dbhandle = 	(int) str_gnum (st[2]),
		  txhandle = 	(int) str_gnum (st[3]),
		  
		  /* Instantiate a new Q structure: */
		  qptr.q = (Query *) malloc(sizeof(Query));
		  
		  if(q_gen(status, qptr.q, t, dbhandle, txhandle)) {
			free(qptr.q);
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], qptr.d);
		  }
	    }
	    break;
	    

	  case US_dbexecute:
	    if(items != 2) {
		  fatal("Usage: $retcode = &dbexecute($qhandle, $txhandle)");
	    } else {

		  /* Get qptr union and txhandle: */
		  qptr.d  = str_gnum (st[1]);
		  txhandle = (int) str_gnum (st[2]);
		  
		  if(q_execute(status, qptr.q, txhandle)) {
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], SUCCESS_VALUE);
		  }
		  
		  free(qptr.q);
	    }
	    break;
		  

	  case US_dbopencursor:
	    if(items != 2) {
		  fatal("Usage: $retcode = &dbopencursor($qhandle, $txhandle)");
	    } else {
		  
		  /* Resolve the q and tx: */
		  qptr.d  = str_gnum (st[1]);
		  txhandle = (int) str_gnum (st[2]);
		  
		  if(q_open_cursor(status, qptr.q, txhandle)) {
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], SUCCESS_VALUE);
		  }
	    }
	    break;



	  case US_dbfetchrow:
	    if(items != 1) {
		  fatal("Usage: @rowdata = &dbfetchrow($qhandle)");
	    } else {
		  /* Resolve the q: */
		  qptr.d = str_gnum (st[1]);
		  
		  j = q_fetch(status, qptr.q, fptrs);
		  
		  /*
		   *  Back up the stack.  If j == Q_EOC (out of records),
		   *  then (return sp) will return NULL and enable us to break
		   *  out of the &dbfetchrow() loop:
		   */
		  --sp;
		  
		  if(j == Q_EOC) {
			break;
		  }

		  for(k = 0; k < qptr.q->sqlda->sqln; k++) {
			
			(void) q_format_fld(s, 0, &fptrs[k]);
			
			(void) astore(ary,++sp,str_2static(str_make(s, 0)));
		  }
	    }
	    break;
		  
		  
		  
	  case US_dbclosecursor:
	    if(items != 1) {
		  fatal("Usage: $retcode = &dbclosecursor($qhandle)");
	    } else {
		  /* Resolve the q: */
		  qptr.d = str_gnum (st[1]);
		  
		  if(q_close_cursor(status, qptr.q)) {
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], SUCCESS_VALUE);
		  }
		  
		  free(qptr.q);
	    }
	    break;
      

	  case US_dbcommit:
	    if(items != 1) {
		  fatal("Usage: $retcode = &dbcommit($txhandle)");
	    } else {
		  /* Resolve the tx: */
		  txhandle = (int) str_gnum (st[1]);

		  if(q_commit_trans(status, &txhandle)) {
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], SUCCESS_VALUE);
		  }
	    }
	    break;
		  

	  case US_dbrollback:
	    if(items != 1) {
		  fatal("Usage: $retcode = &dbrollback($txhandle)");
	    } else {
		  /* Resolve the tx: */
		  txhandle = (int) str_gnum (st[1]);

		  if(q_rollback_trans(status, &txhandle)) {
			str_numset(st[0], ERROR_VALUE);
		  } else {
			str_numset(st[0], SUCCESS_VALUE);
		  }
	    }
	    break;



	  case US_dberrtxt:
	    --sp;    /* Back up the stack 'cause we'll PRE-increment later */

	    /*
	     *  Check status[1].  If -1, then the problem was with dsql so
	     *  process that; otherwise,  Interbase went sour, so use 
	     *  the builtin isc_ function to parse...
	     */
	    if(status[1] == -1) {

		  (void) astore(ary,++sp,str_2static(str_make("interperl error",0)));		  
		  for(i = 2; status[i] != 0; i++) {
			t = (char *) status[i];
			(void) astore(ary,++sp,str_2static(str_make(t, 0)));
		  }
	    } else {
		  int *v;

		  (void) astore(ary,++sp,str_2static(str_make("Interbase error",0)));
		  v = status;
		  while(isc_interprete(s, &v)) {
			(void) astore(ary,++sp,str_2static(str_make(s, 0)));
		  }
	    }
	    break;
	    



	  default:
	    fatal("Unimplemented user-defined subroutine");
      }

      return sp;
}




/* 
 * Return the value of a userdefined variable. These variables are all 
 * READ-ONLY in Perl.
 */
static int
userval(ix, str)
int ix;
STR *str;
{
    return 0;
}

static int
userset(ix, str)  /* Not used. None of these variables are user-settable */
int ix;
STR *str;
{
    return 0;
}
